mkdir -p /etc/gitlab-runner
cat > /etc/gitlab-runner/config.toml <<- EOF

${runners_config}

EOF

${pre_install}

if [[ `echo ${runners_executor}` == "docker" ]]
then
  echo 'installing docker'
  if grep -q ':2$' /etc/system-release-cpe  ; then
    # AWS Linux 2 provides docker via extras only and uses systemd (https://aws.amazon.com/amazon-linux-2/release-notes/)
    amazon-linux-extras install docker
    usermod -a -G docker ec2-user
    systemctl enable docker
    systemctl start docker
  else
    yum install docker -y
    usermod -a -G docker ec2-user
    service docker start
  fi
fi

curl --fail --retry 6 -L https://packages.gitlab.com/install/repositories/runner/gitlab-runner/script.rpm.sh | bash
yum install gitlab-runner-${gitlab_runner_version} -y

if [[ `echo ${docker_machine_download_url}` == "" ]]
then
  curl --fail --retry 6 -L https://github.com/docker/machine/releases/download/v${docker_machine_version}/docker-machine-`uname -s`-`uname -m` >/tmp/docker-machine
else
  curl --fail --retry 6 -L ${docker_machine_download_url} >/tmp/docker-machine
fi

chmod +x /tmp/docker-machine && \
  mv /tmp/docker-machine /usr/local/bin/docker-machine && \
  ln -s /usr/local/bin/docker-machine /usr/bin/docker-machine
docker-machine --version

# Create a dummy machine so that the cert is generated properly
# See: https://gitlab.com/gitlab-org/gitlab-runner/issues/3676
# See: https://github.com/docker/machine/issues/3845#issuecomment-280389178
export USER=root
export HOME=/root
docker-machine create --driver none --url localhost dummy-machine
docker-machine rm -y dummy-machine
unset HOME
unset USER

# Install jq if not exists
if ! [ -x "$(command -v jq)" ]; then
  yum install jq -y
fi

token=$(aws ssm get-parameters --names "${secure_parameter_store_runner_token_key}" --with-decryption --region "${secure_parameter_store_region}" | jq -r ".Parameters | .[0] | .Value")
if [[ `echo ${runners_token}` == "__REPLACED_BY_USER_DATA__" && `echo $token` == "null" ]]
then
  if [[ `echo ${gitlab_runner_registration_token}` != "" ]]
  then
    token=$(curl --request POST -L "${runners_gitlab_url}/api/v4/runners" \
    --form "token=${gitlab_runner_registration_token}" \
    --form "tag_list=${gitlab_runner_tag_list}" \
    --form "description=${giltab_runner_description}" \
    --form "locked=${gitlab_runner_locked_to_project}" \
    --form "run_untagged=${gitlab_runner_run_untagged}" \
    --form "maximum_timeout=${gitlab_runner_maximum_timeout}" \
    --form "access_level=${gitlab_runner_access_level}" \
    | jq -r .token)
    aws ssm put-parameter --overwrite --type SecureString  --name "${secure_parameter_store_runner_token_key}" --value="$token" --region "${secure_parameter_store_region}"
  else
    echo "ERROR: Can't register the runner. The gitlab_runner_registration_token is required on initial deployment to do the agent registration."
  fi
else
  echo "WARNING: Runner already registered. Looking for orphan runners in the runners_security_group_id to destroy it"
  terminate=$(aws ec2 describe-instances --region "${secure_parameter_store_region}" --filter "Name=instance.group-id,Values=${runners_security_group_id}" --query "Reservations[*].Instances[*].InstanceId" --output text)
  if [[ `echo $terminate` != "" ]]
  then
    aws ec2 terminate-instances --instance-ids $terminate --region "${secure_parameter_store_region}"
  fi
fi

sed -i.bak s/__REPLACED_BY_USER_DATA__/`echo $token`/g /etc/gitlab-runner/config.toml

# A small script to remove this runner from being registered with Gitlab. 
cat <<REM > /etc/rc.d/init.d/remove_gitlab_registration
#!/bin/bash
# chkconfig: 1356 99 03
# description: cleans up gitlab runner key
# processname: remove_runner_key
#              /etc/rc.d/init.d/remove_gitlab_registration
lockfile=/var/lock/subsys/remove_gitlab_registration

# This lockfile is necessary so that we'll run the cleanup later.
start() {
    logger "Setting up Runner Removal Lockfile"
    touch \$lockfile
}

# Overwrite token in SSM with null and remove runner from Gitlab
stop() {
    logger "Removing Gitlab Runner Token"
    aws ssm put-parameter --overwrite --type SecureString  --name "${secure_parameter_store_runner_token_key}" --region "${secure_parameter_store_region}" --value="null" 2>&1 | logger &
    curl -sS --request DELETE "${runners_gitlab_url}/api/v4/runners" --form "token=$token" 2>&1 | logger &
    retval=\$?
    rm -f \$lockfile
    return \$retval
}

# Map these to start just to be redunant.
# We don't want to run Stop outside of shutdown.
restart() {
  start
}
reload() {
  start
}

# Do nothing - there's no status.
status() {
  :
}

case "\$1" in
    start)
        \$1
        ;;
    stop)
        \$1
        ;;
    restart)
        \$1
        ;;
    status)
        \$1
        ;;
    *)
        echo "Usage: \$0 {start|stop|status|restart}"
        exit 2
        ;;
esac
REM

chmod a+x /etc/init.d/remove_gitlab_registration

# Use chkconfig to link into the runlevel 0 (shutdown) directories
# This adds "start" scripts to levels 1,3,5, and 6, and a "stop" to the others.
# This way we'll not be assigned jobs if we're shutting down, and clean up in Gitlab.
chkconfig --add remove_gitlab_registration

# As noted above, this does nothing more than make the lockfile.
service remove_gitlab_registration start

${post_install}

service gitlab-runner restart
chkconfig gitlab-runner on
